#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"

extern char *program;

#define SYS_CONF_DIR	"/etc"
#define CONF_NAME_LEN	512
#define BUF_LEN		512

char conf_name[CONF_NAME_LEN];

#define HASH_SIZE	1024

struct key_val {
	char *key;
	char *val;
	struct key_val *next;
};

struct key_val *hash_tab[HASH_SIZE];

static inline struct key_val **hash(char *key) {
	uint32_t val = 0;
	while (*key) {
		val <<= 7 | ((val >> 25) & 0x7f);
		val ^= (uint32_t)(*key) & 0xff;
		key++;
	}
	return &hash_tab[val % HASH_SIZE];
}

static int conf_load_sub(FILE *fp) {
	struct key_val *kv,**head;
	char *top,*val,*p;
	char buf[BUF_LEN];

	while (fgets(buf, BUF_LEN, fp)) {
		p = buf;
		while (*p == ' ' || *p == '\t') p++;
		top = p;
		while (*p && *p != '#' && *p != '\r' && *p != '\n') p++;
		*p = 0;
		p = malloc(strlen(top)+1);
		if (!p) continue;
		strcpy(p, top);
		top = p;
		while (('a' <= *p && *p <= 'z') || ('A' <= *p && *p <= 'Z') 
			|| ('0' <= *p && *p <= '9') || (*p == '.') || (*p == '_') ) p++;
		if (*p == '=') *p++ = 0;
		else  {
			*p++ = 0;
			while (*p && *p != '=') p++;
			if (!*p) continue;
			p++;
		}
		while (*p == ' ' || *p == '\t') p++;
		val = p;
		while (*p && !(*p == ' ' || *p == '\t')) p++;
		*p = 0;
		if (*top && *val) {
			kv = malloc(sizeof(struct key_val));
			if (!kv) continue;
			head = hash(top);
			kv->key = top;
			kv->val = val;
			kv->next = *head;
			*head = kv;
		}
	}
	return 1;
}

static void conf_load() {
	FILE *fp;
	int len;
	char *p;
	len = strlen(program);
	if (len > CONF_NAME_LEN - 10) return;

	strcpy(conf_name, program);
	while (p = strchr(conf_name, '\\')) *p = '/';
	if (!strcasecmp(conf_name+len-4 , ".exe"))
		conf_name[len-4] = 0;
	strcat(conf_name, ".conf");

	fp = fopen(conf_name, "r");
	if (fp) {
		if (conf_load_sub(fp)) return;
	}
	p = strrchr(program, '\\');
	if (!p) p = strrchr(program, '/');
	if (p) p = p+1;
	else p = program;

	strcpy(conf_name, SYS_CONF_DIR);
	strcat(conf_name, "/");
	strcat(conf_name, p);
	len = strlen(conf_name);

	if (!strcasecmp(conf_name+len-4 , ".exe"))
		conf_name[len-4] = 0;
	strcat(conf_name, ".conf");

	fp = fopen(conf_name, "r");
	if (fp) {
		if (conf_load_sub(fp)) return;
	}
	conf_name[0] = 0;
}

char *conf_getenv(char *name) {
	struct key_val *p;
	if (!conf_name[0]) {
		conf_load();
	}
	p = *hash(name);
	while (p) {
		if (!strcmp(name, p->key)) return p->val;
		p = p->next;
	}
	return NULL;
}
