#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "jtag.h"
#include "target.h"

//#define ISP_CHAN	2 : moved to target.h

//#define ISP_SUPPORT_SIN_SOUT
#define ISP_TEST

// SIN       0aa1 aaaa  XXXX XXXX
//           XXXX XXXX  load_data
//
// SOUT      1aa1 aaaa  store_data
//           XXXX XXXX  XXXX XXXX
//
// SSTPR     0110 100a  PR[a]
//           XXXX XXXX  XXXX XXXX
//
// SLD       0010 0I00  XXXX XXXX    ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  load_data
//
// SST       0110 0I00  store_data   ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  XXXX XXXX

/* TPI instructions */
#define TPI_SLD_POSTI 0x24 // 
#define TPI_SST_POSTI 0x64 // 
#define TPI_SETPR_LO  0x68 // 
#define TPI_SETPR_HI  0x69 // 
#define ISP_TPI_SIN   0x10 // 
#define TPI_SOUT      0x90 // 

#define NVMCSR  0x32
#define NVMCSR_BUSY 0x80

#define NVMCMD 0x33
#define NVMCMD_NOP   0x00
#define NVMCMD_WRITE 0x1d

/* TPI operations */
static uint16_t tpi_addr;

static int tpiSetPR(CABLE cbl, uint16_t addr) {
	uint32_t s;
	s = 0x6800 | (addr & 0xff); // TPI_SETPR_LO
	jtag_SDR32R(cbl, s, 16, JTAG_CONT, NULL);

	s = 0x6900 | ( ( addr >> 8 ) & 0xff); // TPI_SETPR_HI
	jtag_SDR32R(cbl, s, 16, JTAG_CONT, NULL);

	return 0;
}

static int tpiReadDS(CABLE cbl, uint16_t address) {
	uint32_t s,r;
	if (address != tpi_addr) {
		tpiSetPR(cbl, address);
		tpi_addr = address;
	}
	s = 0x2400; // TPI_SLD_POSTI
	jtag_SDR32R(cbl, s, 16, JTAG_RECV|JTAG_CONT, &r);
	tpi_addr ++;
	return (r & 0xff);
}

static int tpiWriteDS(CABLE cbl, uint16_t address, uint8_t data) {
	uint32_t s;
	if (address != tpi_addr) {
		tpiSetPR(cbl, address);
		tpi_addr = address;
	}
	s = 0x6400 | data; // TPI_SST_POSTI
	jtag_SDR32R(cbl, s, 16, JTAG_CONT, NULL);
	tpi_addr ++;
	return 0;
}

static int tpiReadCS(CABLE cbl, unsigned int address) {
#ifdef ISP_SUPPORT_SIN_SOUT
	uint8_t inst;
	int s,r;
	inst = address & 0xf;
	if (address & 0x8000) { /* read CS */
		inst |= 0x80; // TPI_SLDCS
	} else { /* read IO */
		inst |= (address << 1) & 0x60;	
		inst |= 0x10; // TPI_SIN
	}
	s = (inst << 8);

	jtag_SDR32R(cbl, r, 16, JTAG_RECV|JTAG_CONT, &r);
	return r & 0xff;
#else
	return tpiReadCS(cbl,address);
#endif
}

static int tpiWriteCS(CABLE cbl, unsigned int address, uint8_t data) {
#ifdef ISP_SUPPORT_SIN_SOUT
	uint8_t inst;
	int s;
	inst = address & 0xf;
	if (address & 0x8000) { /* write CS */
		inst |= 0xc0; // TPI_SSTCS
	} else {
		inst |= (address << 1) & 0x60;
		inst |= 0x90; // TPI_SOUT
	}
	s = (inst << 8) | data ;
	jtag_SDR32R(cbl, s, 16, JTAG_CONT, NULL);
	return 0;
#else
	return tpiWriteCS(cbl,address, data);
#endif
}

int isp_program(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int i;
	//int r;
	int addr;
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, ISP_CHAN);

	if (len > 0x2000) len = 0x2000;
	if (len == 0) len = 0x2000;
#ifndef ISP_TEST
	r = tpiReadCS(cbl, NVMCSR);
	tpiWriteCS(cbl, NVMCMD, NVMCMD_WRITE);
	r = tpiReadCS(cbl, NVMCSR);
	if (r & NVMCSR_BUSY) {
		return -1;
	}
#endif
	tpi_addr = 0;
	addr = 0x4000;
	progress_start("Programming", len);
	for (i=0; i<len; i++) {
		tpiWriteDS(cbl, addr, buf[i]);
		addr++;
		progress(i);
	}
#ifndef ISP_TEST
	tpiWriteCS(cbl, NVMCMD, NVMCMD_NOP);
#endif
	tpiReadDS(cbl, 0x04000); // for flush
	progress_stop();
	return 0;
}

int isp_read(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int i;
	int addr;
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, ISP_CHAN);

	if (len > 0x2000) len = 0x2000;
	if (len == 0) len = 0x2000;
	progress_start("Reading", len);
	tpi_addr = 0;
	addr = 0x4000;
	for (i=0; i<len; i++) {
		buf[i] = tpiReadDS(cbl, addr);
		addr++;
		progress(i);
	}
	progress_stop();
	return i;
}
