/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//`define ISP_SUPPORT_SIN_SOUT
`define ISP_TEST
`define USE_DEBUG_LED

// SIN       0aa1 aaaa  XXXX XXXX
//           XXXX XXXX  load_data
// 
// SOUT      1aa1 aaaa  store_data
//           XXXX XXXX  XXXX XXXX
// 
// SSTPR     0110 100a  PR[a]        ( PR[0] : low  byte of PR )
//           XXXX XXXX  XXXX XXXX    ( PR[1] : high byte of PR )
// 
// SLD       0010 0I00  XXXX XXXX    ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  load_data
// 
// SST       0110 0I00  store_data   ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  XXXX XXXX
//
module isp (
		input CLK
		, input MOSI
		, input SCK
		, input CS
		, output MISO

`ifdef USE_DEBUG_LED
		, output [3:0] ISP_STATE
`endif
`ifdef ISP_TEST
`else
		, output ISP_STORE
		, output ISP_LOAD
		, output [15:0] ISP_ADDR
		, output [7:0] ISP_STORE_DATA
		, input [7:0] ISP_LOAD_DATA
`endif
		);

  wire [7:0] SPI_DATA;
  wire [7:0] SPI_DATA_I;
  wire [1:0] SPI_COUNT;
  jtag_spi JTAG_SPI (.CLK(CLK)
		, .MOSI(MOSI)
		, .SCK(SCK)
		, .CS(CS)
		, .MISO(MISO)
		, .SPI_DATA(SPI_DATA)
		, .SPI_DATA_VALID(SPI_DATA_VALID)
		, .SPI_DATA_REQ(SPI_DATA_REQ)
		, .SPI_DATA_COUNT(SPI_COUNT)
		, .SPI_DATA_I(SPI_DATA_I)
		);

`ifdef USE_DEBUG_LED
   reg [3:0] r_state = 0;
   assign ISP_STATE[3:0] = r_state[3:0];
`endif
`ifdef ISP_TEST
		wire ISP_STORE;
		wire ISP_LOAD;
		wire [15:0] ISP_ADDR;
		wire [7:0] ISP_STORE_DATA;
		wire [7:0] ISP_LOAD_DATA;
`endif


`ifdef LATCHED_INPUT
    reg [7:0] r_cmd;
`else
    wire [7:0] r_cmd = SPI_DATA;
`endif
    reg r_req;
    reg r_valid;
    reg r_sstpr_lo;
    reg r_sstpr_hi;
    reg r_pr_inc;
    reg r_store;
    reg r_pr_mem;
    reg [15:0] r_pr;
    reg [7:0] r_data_out;
    
`ifdef ISP_SUPPORT_SIN_SOUT
    assign ISP_ADDR[15:0] = r_sin_sout ? { 10'h000, r_abs } : r_pr;
`else
    assign ISP_ADDR[15:0] = r_pr;
`endif

`ifdef ISP_SIMPLE_DECODE
    wire f_sstpr    = (r_cmd[7:5] == 3'b011) & r_cmd[3];
    wire f_sstpr_hi = f_sstpr &  r_cmd[0];
    wire f_sstpr_lo = f_sstpr & ~r_cmd[0];
    wire f_sld        =  (r_cmd[7:5] == 3'b001) & ~r_cmd[3]; 
    wire f_sst        =  (r_cmd[7:5] == 3'b011) & ~r_cmd[3];
    wire f_pr_inc     = r_cmd[2];
`else
  `ifdef ISP_SUPPORT_SIN_SOUT
// SIN       0aa1 aaaa  XXXX XXXX
// SOUT      1aa1 aaaa  store_data
    wire f_sin_sout = r_cmd[4];
    wire f_sin      = f_sin_sout & ~r_cmd[7];
    wire f_sout     = f_sin_sout &  r_cmd[7];
  `endif
// SSTPR     0110 100a  PR[a]
    wire f_sstpr    = (r_cmd[7:1] == 7'b0110100);
    wire f_sstpr_hi = f_sstpr &  r_cmd[0];
    wire f_sstpr_lo = f_sstpr & ~r_cmd[0];
// SLD       0010 0I00  XXXX XXXX    ( if I == 1 , PR = PR +1 )
// SST       0110 0I00  store_data   ( if I == 1 , PR = PR +1 )
//           * ** * **
    wire f_sld_sst    = ~r_cmd[7] & ( r_cmd[5:3] == 3'b100 ) & (r_cmd[1:0] == 2'b00 );
    wire f_sld        = f_sld_sst & ~r_cmd[6];
    wire f_sst        = f_sld_sst &  r_cmd[6];
    wire f_pr_inc     = f_sld_sst &  r_cmd[2];

  `ifdef ISP_SUPPORT_SIN_SOUT
    wire f_store = f_sout | f_sst;
    wire f_load  = f_sin  | f_sld;

    wire [5:0] v_abs = { 10'h000, r_cmd[6:5], r_cmd[3:0] }
  `else
    wire f_store = f_sst;
    wire f_load  = f_sld;
  `endif
`endif
    assign r_load = ( r_pr_inc & ~r_store );
    assign SPI_DATA_I[7:0]      = r_data_out;
    assign ISP_STORE = (SPI_DATA_VALID & ~r_valid & SPI_COUNT[0]) & r_store;
    assign ISP_STORE_DATA[7:0] = SPI_DATA;
   
    always @(posedge CLK)
    begin
	if (CS)
    	    begin
		r_req          <= 1'b0;
		r_valid        <= 1'b0;
		r_pr_inc       <= 1'b0;
		r_pr_mem       <= 1'b0;
		r_sstpr_hi     <= 1'b0;
		r_sstpr_lo     <= 1'b0;
		r_store        <= 1'b0;
    	    end
	else
	  begin
	    r_req    <= SPI_DATA_REQ;
	    r_valid  <= SPI_DATA_VALID;
	    if (SPI_DATA_REQ & SPI_COUNT[0])
    	      begin
		r_data_out <= ISP_LOAD_DATA;
	      end
	    else if (SPI_DATA_REQ & ~SPI_COUNT[0])
    	      begin
		r_data_out[0] <= r_pr_inc;
		r_data_out[1] <= r_load;
		r_data_out[2] <= r_store;
		r_data_out[3] <= 1'b1;
		r_data_out[7:4] <= 4'hA;
	      end
	    if (SPI_DATA_VALID  & ~r_valid & ~SPI_COUNT[0])
    	      begin
		r_pr_inc    <= f_pr_inc;
		r_sstpr_hi  <= f_sstpr_hi;
		r_sstpr_lo  <= f_sstpr_lo;
		r_store     <= f_store;
                r_pr_mem   <= (ISP_ADDR[15:12] == 4'b0100);
	      end
	    if (SPI_DATA_VALID  & ~r_valid  & SPI_COUNT[0])
    	      begin
		if (r_pr_inc) r_pr <= r_pr + 1;
		else if (r_sstpr_hi) r_pr[15:8] <= ISP_STORE_DATA;
		else if (r_sstpr_lo) r_pr[7:0]  <= ISP_STORE_DATA;
	      end
	 end
	 
`ifdef USE_DEBUG_LED
	if (CS)
    	    begin
		r_state       <= 0;
    	    end
	else
	  begin
	    if (r_req & ~SPI_COUNT[0])
    	      begin
		r_state[0] <= f_pr_inc;
		r_state[1] <= f_load;
		r_state[2] <= f_store;
	        r_state[3] <= 1'b0;
	      end
	    else if (r_req &  SPI_COUNT[0])
	        r_state[3] <= 1'b1;
	 end
`endif
    end

`ifdef ISP_TEST
//output ISP_STORE
//output ISP_LOAD
//output [15:0] ISP_ADDR
//output [7:0] ISP_STORE_DATA
//input [7:0] ISP_LOAD_DATA

    reg  [7:0] mem [0:2048-1];
    reg  [10:0] r_addra;

    always @(negedge CLK)
        begin
                if(ISP_STORE & r_pr_mem)
			 mem[ISP_ADDR[10:0]] <= ISP_STORE_DATA;
                r_addra <= ISP_ADDR[10:0];
        end
    assign ISP_LOAD_DATA[7:0] = mem[r_addra];

  initial
    begin
       $readmemh("rom_data.mem" , mem, 0 , 2048-1);
    end
`endif
endmodule
